module ActiveQuery
  module SQLParser

    require 'strscan'

    class InvalidQueryError < ArgumentError; end
    class SyntaxError < StandardError
      attr_reader :line, :pos
      def initialize(msg = "", line:, pos:)
        msg = "Missing #{msg} on line #{line} (pos. #{pos})"
        super(msg)
      end
    end

    Query = Struct.new(:name, :params, :comment, :query) 
  
    WHITESPACE = /\s+/
    NAME = /^\s*--\s*name\s*:\s*(.+)/
    QNAME = /^\s*--\s*name\s*:/
    SEMICOLON = /;/
    PARAMS = /\(\:\s*([^)]+?)\s*\)/
    ALL_COMMENTS = /(?-m:--.*)|(?m:--\*.--?\*\/)/
    PARAM = /:\w+/
   
    def self.parse_query(query)
      @scanner = StringScanner.new(query.strip)
      @line = 0
      @query = Query.new 

      until @scanner.eos?
        @line += 1
        parse_line
      end
      @query.query = make_params(@query.query, @query.params)
      @query
    end

    private
    def self.parse_line
      @query.params = []
      # Query name match: --name:
      header = @scanner.check_until(NAME)
      unless header
         fail SyntaxError.new("'-- name:'", line: @line, pos: @scanner.pos)
      end

      ## ===============================
      # Query name match after: --name:
      # Todo: test missing name 
      query_name = @scanner.scan_until(NAME)
      unless query_name 
        fail SyntaxError.new("query name", line: @line, pos: @scanner.pos)        
      end
      @query.name = @scanner.captures&.first.strip

      # SQL expression
      @scanner.skip_until(ALL_COMMENTS)
      @scanner.skip_until(/^\s+/)
      query = @scanner.check_until(SEMICOLON)
      if query && query != ';'
        @query.query = query.strip
      else
         fail SyntaxError.new("SQL expression", line: @line, pos: @scanner.pos)
      end

      # All params 
      until @scanner.check_until(PARAM).nil?
        @scanner.scan_until(PARAM)
        @query.params.push @scanner.matched
      end
      # done 
      @scanner.scan_until(SEMICOLON) # important! don't touch
    end

    # and this only work for postgres - split transform from struct
    def self.make_params(query, params)
      if params && query
        params.each_with_index do |param, i| 
          pos = "$#{i+1}"
          query.gsub!(param, pos) 
        end
        query.strip if query
      else
        query.strip if query
      end
    end

    def self.valid_query?(query)
      query.name && query.name != "" && 
      query.query && query.query != ""
    end

  end
end

